/*{{{  include files*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <math.h>

#include "fftl2.h"
#include "tlmproc.h"
/*}}}  */
/*{{{  structures*/
/* ROW telemetry data file logical record header */
struct row_file_header {
  unsigned char tag, chan;         /* Always 0, 1 */
  long time, size;
  };
/*}}}  */
/*{{{  global variables*/
static unsigned char HKblock[HK_SIZE], SCIblock[SCI_SIZE];
double OnBoardTimes[8], SnapshotOffsets[8];

int SSoffset_mode = 1;
extern int Verbosity;
/*}}}  */
/*{{{  struct tm *decodetime (timep)*/
struct tm *decodetime (timep)
long *timep;
{
static struct tm timebuf;
struct tm *gmptr;

if ((gmptr = gmtime (timep)) != NULL) return (gmptr);

timebuf.tm_year = 0;
timebuf.tm_mon  = 0;
timebuf.tm_mday = 0;
timebuf.tm_hour = (int)((*timep / 3600) % 24);
timebuf.tm_min  = (int)((*timep /   60) % 60);
timebuf.tm_sec  = (int)(*timep % 60);
return (&timebuf);
}

/*}}}  */
/*{{{  char *doubledate (dtime)*/
char *doubledate (dtime)
double dtime;
{
struct tm *inttime;
static char strtime[64];
long ltime;

ltime = (long)dtime;
inttime = decodetime (&ltime);
if (inttime == NULL) return ("****/**/** **:**:**.******");
sprintf (strtime, "%4d/%02d/%02d %02d:%02d:%02d.%06.0f",
         inttime->tm_year+1900, inttime->tm_mon+1, inttime->tm_mday,
         inttime->tm_hour, inttime->tm_min, inttime->tm_sec,
         (dtime - floor (dtime)) * 1000000.0);
return (strtime);
}
/*}}}  */
/*{{{  char *doubletime (dtime)*/
char *doubletime (dtime)
double dtime;
{
struct tm *inttime;
static char strtime[16];
long ltime;

ltime = (long)dtime;
inttime = decodetime (&ltime);
if (inttime == NULL) return ("**:**:**");
sprintf (strtime, "%02d:%02d:%02d",
         inttime->tm_hour, inttime->tm_min, inttime->tm_sec);
return (strtime);
}
/*}}}  */
/*{{{  long encode_time (string)*/
long encode_time (string)
char *string;
{
int hour, min, sec;

sscanf (string, "%d:%d:%d", &hour, &min, &sec);
return ((long)hour * 3600 + (long)min * 60 + sec);
}
/*}}}  */
/*{{{  long read_lrec (filep, bufsize, format, bufptr, timeptr)*/
long read_lrec (filep, bufsize, format, bufptr, timeptr)
FILE *filep;
long bufsize;
int format;
unsigned char  *bufptr;
double *timeptr;
/* Reads one logical record from binary telemetry file. Returns size of
   record found (bytes) or -1 for end of file, -2 for error. */
{
long tlm_blk_size;
struct row_file_header head;
unsigned char SIShead[SIS_HEADSIZE];
unsigned char DDShead[DDS_HEADSIZE];

switch (format) {
  case fmROW_TLM:
  /*{{{  read ROW format file*/
  if (fread (&head, 10, 1, filep) < 1) return ((long)-1);
  /* printf ("read_lrec: size = %d\n", (int)(head.size / 2)); */
  if (head.tag != 0 || head.chan != 1 ||
      head.size < 0 || head.size > bufsize) return ((long)-2);
  if (timeptr != NULL) *timeptr = head.time * 0.001;
  if (fread (bufptr, (size_t)(head.size), 1, filep) < 1) return (-2);
  tlm_blk_size = head.size;
  break;
  /*}}}  */
  case fmCCS_TLM:
  case fmSIS_TLM:
  case fmFDG_TLM:
  /*{{{  read SIS/CCS format file*/
  if (fread (SIShead, SIS_HEADSIZE, 1, filep) < 1) return ((long)-1);
  tlm_blk_size = (long)(SIShead[4]) * 256 + SIShead[5] - 7;
  if (tlm_blk_size > bufsize) return ((long)-2);
  if (fread (bufptr, (size_t)tlm_blk_size, 1, filep) < 1) return ((long)-2);
  *timeptr = SIShead[7] * 256.0*65536.0 + SIShead[8] * 65536.0 +
             SIShead[9] * 256.0 + SIShead[10] * 1.0 + SIShead[11] / 256.0 +
             SIShead[12] / 65536.0 + SIShead[13] / (256.0*65536.0);
  /*}}}  */
  break;
  case fmDDS_TLM:
  /*{{{  read DDS format file*/
  if (fread (DDShead, DDS_HEADSIZE, 1, filep) < 1) return ((long)-1);
  tlm_blk_size = (long)(DDShead[10]) * 256 + DDShead[11];
  if (tlm_blk_size > bufsize) return ((long)-2);
  if (fread (bufptr, (int)tlm_blk_size, 1, filep) < 1) return ((long)-2);
  *timeptr = (DDShead[0] * 256.0 + DDShead[1] - 4383) * 86400.0 +
             DDShead[2] * 16777.216 + DDShead[3] * 65.536 +
             DDShead[4] * 0.256 + DDShead[5] * 0.001 +
             DDShead[6] * 0.000256 + DDShead[7] * 0.000001;
  /*}}}  */
  break;
  default:
  return ((long)-2);
  }
/*
fprintf (stderr, "Read block of %ld bytes at %16.6f\n", tlm_blk_size, *timeptr);
*/
return (tlm_blk_size);
}

/*}}}  */
/*{{{  static int proc_sci_pkt (tlm_ptr, pkt_size)*/
static int proc_sci_pkt (tlm_ptr, pkt_size)
unsigned char *tlm_ptr;
int pkt_size;
{
return (proc_sci (tlm_ptr, pkt_size,
        OnBoardTimes[reset_cnt (tlm_ptr)],
        SnapshotOffsets[reset_cnt (tlm_ptr)]));
}
/*}}}  */
/*{{{  static void asmb_sci_pkt (tlm_ptr, pkt_size, sci_time)*/
static void asmb_sci_pkt (tlm_ptr, pkt_size, sci_time)
unsigned char *tlm_ptr;
int pkt_size;
double sci_time;
{
/*{{{  local variables and tables*/
int i;
static unsigned char AsmbBuffer[MAXASMBSIZE];
static int AsmbSize = 0, BlockCount = 0, AsmbFlag = 0;
/*}}}  */

/*{{{  check for partially assembled block at EOF*/
if (pkt_size == 0) {  /* Special call to handle EOF */
  if (AsmbFlag) {
    write_err (sci_time, "EOF found or error reading file during re-assembly");
    }
  return;
  }
/*}}}  */

/*{{{  check for invalid block for re-assembly*/
if (AsmbFlag &&
    (blkpos(tlm_ptr) == SINGLE_BLK || blkpos(tlm_ptr) == FIRST_BLK ||
     source(tlm_ptr) != source(AsmbBuffer) || pkt_size + AsmbSize > MAXASMBSIZE+1)) {
  proc_sci_pkt (AsmbBuffer, AsmbSize);
  if (Verbosity & vDECOM) {
    write_err (sci_time, "Incorrect block found during re-assembly");
    printf ("... blkpos = %d, source = %d, size = %d+%d\n",
              blkpos(tlm_ptr), source(tlm_ptr), pkt_size, AsmbSize);
  }
  AsmbFlag = FALSE;
  AsmbSize = 0;
  }
/*}}}  */

/*{{{  assemble subsequent blocks and return*/
if (AsmbFlag) {
  if (blkcount(tlm_ptr) != BlockCount && (Verbosity & vDECOM))
     write_err (sci_time, "One or more sub-blocks appear to be missing");
  BlockCount = (blkcount(tlm_ptr) + 1) % 4;
  for (i = PKT_HEADSIZE; i < pkt_size; i++) AsmbBuffer[AsmbSize + i - PKT_HEADSIZE] = tlm_ptr[i];
  AsmbSize += pkt_size - PKT_HEADSIZE;
  if (blkpos(tlm_ptr) == LAST_BLK) {
    proc_sci_pkt (AsmbBuffer, AsmbSize);
    AsmbFlag = FALSE;
    AsmbSize = 0;
    }
  return;
  }
/*}}}  */

/*{{{  check for invalid block not during re-assembly*/
if (!AsmbFlag && (blkpos(tlm_ptr) == MIDDLE_BLK || blkpos(tlm_ptr) == LAST_BLK))
  {
  if (Verbosity & vDECOM) {
    write_err (sci_time, "Middle or last block when single or first expected");
    printf ("... blkpos = %d, source = %d, size = %d+%d\n",
          blkpos(tlm_ptr), source(tlm_ptr), pkt_size, AsmbSize);
  }
  }
/*}}}  */

/*{{{  start a re-assembly sequence and return*/
if (blkpos(tlm_ptr) == FIRST_BLK && pkt_size <= MAXASMBSIZE) {
  BlockCount = 0;
  for (i = 0; i < pkt_size; i++) AsmbBuffer[i] = tlm_ptr[i];
  AsmbFlag = TRUE;
  AsmbSize = pkt_size;
  return;
  }
/*}}}  */

/*{{{  handle individual blocks*/
proc_sci_pkt (tlm_ptr, pkt_size);
/*}}}  */
}
/*}}}  */
/*{{{  double CalcSSoffset (hk_ptr)*/
double CalcSSoffset (hk_ptr)
unsigned char *hk_ptr;
{
return 0.0;
}
/*}}}  */

void UpdateOBT (int reset_count, double hk_time)
{
	int i, obt_index;
	double extrap_time;

	/* Update the primary entry */
	OnBoardTimes[reset_count % 8] = hk_time;

	/* Update any preceding entries that are zero */
	for (i = -7; i < 0; i++) {
		extrap_time = hk_time + i * 5.15222168;
		obt_index = (reset_count + i + 8) % 8;
		if (OnBoardTimes[obt_index] == 0.0) OnBoardTimes[obt_index] = extrap_time;
	}
}

/*{{{  int proc_file (hk_file, sci_file, start, finish, format)*/
int proc_file (hk_file, sci_file, start, finish, format)
FILE *hk_file, *sci_file;
long start, finish;
int format;
{
/*{{{  local variables*/
unsigned char *tlm_data_ptr;
long HK_blk_size, SCI_blk_size, time_of_day;
int ResetCount, inst_blk_size, inst_blk_pos;
int tlm_blk_size, SCI_num_blks, blk, dif_blk_cnt;
static double HKtime, SCItime;
/*}}}  */

SCItime = HKtime = 0.0;
dif_blk_cnt = 0;
while ((HK_blk_size = read_lrec (hk_file, (long)HK_SIZE, format, HKblock, &HKtime)) >= 0) {
  dif_blk_cnt--;
  while (HKtime - 6.152 > SCItime) {
    if ((SCI_blk_size = read_lrec (sci_file, (long)SCI_SIZE,
                                   format, SCIblock, &SCItime)) < 0) {
      printf ("proc_file: status %ld on science file.\n", SCI_blk_size);
      return (SCI_blk_size == -1);
      }
    dif_blk_cnt++;
    }
  /*{{{  record time info*/
  ResetCount = hk_rst_cnt (HKblock);
  UpdateOBT (ResetCount, HKtime);
  SnapshotOffsets[(ResetCount+7) % 8] = CalcSSoffset (HKblock);
  time_of_day = (long) HKtime % 86400;
  /*}}}  */
  if (HKtime - 4.152 > SCItime) {
    if (time_of_day >= start && proc_hk (HKblock, HK_SIZE, HKtime)) {
      /*{{{  check if resync has occurred*/
      if (dif_blk_cnt != 0 && (Verbosity & vDECOM)) {
        printf ("Resync: (%s) %d excess science blocks read.\n",
                                    doubletime (HKtime), dif_blk_cnt);
        dif_blk_cnt = 0;
        }
      /*}}}  */
      if (SCI_blk_size >= 306*62) SCI_num_blks = 62;
      else if (SCI_blk_size >= 1000) SCI_num_blks = 10;
      else SCI_num_blks = 1;
      for (blk = 0; blk < SCI_num_blks; blk++) {
        /*{{{  process each spacecraft block*/
        tlm_blk_size = (unsigned)(SCI_blk_size / SCI_num_blks);
        tlm_data_ptr = SCIblock + tlm_blk_size * blk;
        proc_raw (tlm_data_ptr, tlm_blk_size, SCItime);
        inst_blk_size = blksize (tlm_data_ptr);
        inst_blk_pos  = blkpos (tlm_data_ptr);
        while (inst_blk_size > PKT_HEADSIZE || inst_blk_pos == FIRST_BLK) {
          if (headword (tlm_data_ptr) == 0xffff) break;
          if (inst_blk_size <= tlm_blk_size) {
            asmb_sci_pkt (tlm_data_ptr, inst_blk_size, SCItime);
            }
          else   if (Verbosity & vDECOM) {
             write_err (SCItime, "Instrument block overruns telemetry block");
          }
         tlm_blk_size -= inst_blk_size;
          if (tlm_blk_size > 0) {
            tlm_data_ptr += inst_blk_size;
            inst_blk_size = blksize (tlm_data_ptr);
            inst_blk_pos  = blkpos (tlm_data_ptr);
            }
          else {
            inst_blk_size = 0;
            inst_blk_pos  = SINGLE_BLK;
            }
          }
        if (tlm_blk_size > 0) proc_unused (tlm_data_ptr, tlm_blk_size, SCItime);
        /*}}}  */
        }
      }
    if (time_of_day > finish) return (1);
    }
  }
printf ("proc_file_def: status %ld on HK file.\n", HK_blk_size);
return (HK_blk_size == -1); /* TRUE if eof, FALSE if other error */
}
/*}}}  */

